#include <math.h>
#include "efuns.h"
#include "list.h"
#include "array.h"
#include "main.h"
#include "stralloc.h"
#include "opcodes.h"
#include "simulate.h"

INLINE int is_eq(struct svalue *a, struct svalue *b)
{
  if (a->type != b->type) return 0;
  switch(a->type)
  {
  case T_REGEXP:
  case T_NUMBER:
  case T_LIST:
  case T_POINTER:
  case T_OBJECT:
  case T_MAPPING:
  case T_ALIST_PART:
    return a->u.number == b->u.number;

  case T_STRING:
    return my_string_is_equal(a,b);

  case T_FUNCTION:
    return (a->subtype==b->subtype && a->u.ob==b->u.ob);
      
  case T_FLOAT:
    return a->u.fnum == b->u.fnum;

  default:
    fatal("Unknown type %x\n",a->type);
    return 0; /* make gcc happy */
  }
}

int is_equal(struct svalue *a,struct svalue *b)
{
  struct svalue *ap,*bp;
  int e;

  if(a->type!=b->type) return 0;
  if(is_eq(a,b)) return 1;
  switch(a->type)
  {
    case T_REGEXP:
    case T_NUMBER:
    case T_STRING:
    case T_OBJECT:
    case T_FLOAT:
    case T_FUNCTION:
      return 0;

    case T_POINTER:
    case T_ALIST_PART:
    case T_MAPPING:
    case T_LIST:
    if((e=a->u.vec->size)!=b->u.vec->size) return 0;
    ap=a->u.vec->item;
    bp=b->u.vec->item;
    for(;e>0;e--) if(!is_equal(ap++,bp++)) return 0;
      break;

    default:
      fatal("Unknown type in is_equal.\n");
  }
  return 1; /* survived */
}

INLINE int is_gt(struct svalue *a, struct svalue *b)
{
  if (a->type != b->type)
  {
    error("Cannot compare different types.\n");
  }
  switch(a->type)
  {
  default:
    error("Bad argument 1 tp '<'.\n");

  case T_NUMBER:
    return a->u.number > b->u.number;

  case T_STRING:
    return my_strcmp(a,b)>0;

  case T_FLOAT:
    return a->u.fnum > b->u.fnum;
  }
}

INLINE int is_lt(struct svalue *a, struct svalue *b)
{
  if (a->type != b->type)
  {
    error("Cannot compare different types.\n");
  }
  switch(a->type)
  {
  default:
    error("Bad arg 1 to '<='.\n");

  case T_NUMBER:
    return a->u.number < b->u.number;

  case T_STRING:
    return my_strcmp(a, b)<0;

  case T_FLOAT:
    return a->u.fnum < b->u.fnum;

  }
}

INLINE void f_eq()
{
  int i;
  i=is_eq(sp-1,sp);
  pop_stack();
  pop_stack();
  push_number(i);
}

INLINE void f_ne()
{
  int i;
  i=!is_eq(sp-1,sp);
  pop_stack();
  pop_stack();
  push_number(i);
}

INLINE void f_lt()
{
  int i;
  i=is_lt(sp-1,sp);
  pop_stack();
  pop_stack();
  push_number(i);
}

INLINE void f_gt()
{
  int i;
  i=is_gt(sp-1,sp);
  pop_stack();
  pop_stack();
  push_number(i);
}

INLINE void f_ge()
{
  int i;
  i=!is_lt(sp-1,sp);
  pop_stack();
  pop_stack();
  push_number(i);
}

INLINE void f_le()
{
  int i;
  i=!is_gt(sp-1,sp);
  pop_stack();
  pop_stack();
  push_number(i);
}

void f_sum(int num_arg,struct svalue *argp)
{
  int e,type,size;
  type=argp[0].type;
  if(type==T_OBJECT || type==T_FUNCTION) bad_arg(1,F_SUM);

  for(e=1;e<num_arg;e++)
  {
    if(type!=argp[e].type)
    {
      type=T_STRING;
      break;
    }
  }
  for(e=0;e<num_arg;e++)
  {
    if(argp[e].type!=type &&
       argp[e].type!=T_STRING &&
       argp[e].type!=T_INT &&
       argp[e].type!=T_FLOAT)
      bad_arg(e+1,F_SUM);
  }
  switch(type)
  {
    case T_STRING:
    {
      char *buf,buffer[40];
      for(size=e=0;e<num_arg;e++)
      {
        switch(argp[e].type)
        {
          case T_STRING:
            size+=my_strlen(argp+e);
            break;

          case T_INT:
            size+=14;
            break;

          case T_FLOAT:
            size+=22;
            break;
        }
      }
      buf=malloc(size+1);
      size=0;
      for(size=e=0;e<num_arg;e++)
      {
        switch(argp[e].type)
        {
          case T_STRING:
            MEMCPY(buf+size,argp[e].u.string,my_strlen(argp+e));
            size+=my_strlen(argp+e);
            break;

          case T_INT:
            sprintf(buffer,"%d",argp[e].u.number);
            MEMCPY(buf+size,buffer,strlen(buffer));
            size+=strlen(buffer);
            break;

          case T_FLOAT:
            sprintf(buffer,"%f",argp[e].u.fnum);
            MEMCPY(buf+size,buffer,strlen(buffer));
            size+=strlen(buffer);
            break;
        }
      }
      buf[size]=0;
      pop_n_elems(num_arg);
      push_shared_string(make_shared_binary_string(buf,size));
      free(buf);
    }
    break;

    case T_INT:
    for(size=e=0;e<num_arg;e++) size+=argp[e].u.number;
    pop_n_elems(num_arg);
    push_number(size);
    break;

    case T_FLOAT:
    {
      float sum;
      sum=0.0;
      for(e=0;e<num_arg;e++) sum+=argp[e].u.fnum;
      pop_n_elems(num_arg);
      push_float(sum);
    }
    break;

    case T_POINTER:
    {
      struct vector *v;
      struct svalue *r;
      int d;
      for(size=e=0;e<num_arg;e++)
      {
	check_vector_for_destruct(argp[e].u.vec);
        size+=argp[e].u.vec->size;
      }
      if(argp[0].u.vec->ref==1)
      {
	e=argp[0].u.vec->size;
	v=resize_array(argp[0].u.vec,size);
	argp[0]=const0;
	size=e;
	e=1;
      }else{
	v=allocate_array(size);
	e=size=0;
      }
      for(;e<num_arg;e++)
      {
	r=argp[e].u.vec->item;
	for(d=0;d<argp[e].u.vec->size;d++)
	{
	  assign_svalue_no_free(v->item+size,r+d);
	  size++;
	}
      }
      pop_n_elems(num_arg);
      push_vector(v);
      v->ref--;
    }
    break;

    case T_LIST:
    {
      struct vector *v,*w;
      struct svalue *r;
      int d;
      for(size=e=0;e<num_arg;e++)
      {
	check_alist_for_destruct(argp[e].u.vec);
        size+=argp[e].u.vec->size;
      }
      if(num_arg==2)
      {
	w = do_array_surgery(argp[0].u.vec,argp[1].u.vec,OPER_ADD,T_LIST);
      }else{
        v=allocate_array(size);
        for(size=e=0;e<num_arg;e++)
        {
          r=argp[e].u.vec->item[0].u.vec->item;
          for(d=0;d<argp[e].u.vec->size;d++)
          {
            assign_svalue_no_free(v->item+size,r+d);
            size++;
          }
        }
        w=allocate_array(1);
        w->item[0].u.vec=v;
        w->item[0].type=T_ALIST_PART;
        order_alist(w);
      }
      pop_n_elems(num_arg);
      push_list(w);
      w->ref--;
    }
    break;

    case T_MAPPING:
    {
      struct vector *v,*values,*w;
      struct svalue *r,*rr;
      int d;
      for(size=e=0;e<num_arg;e++)
      {
	check_alist_for_destruct(argp[e].u.vec);
        size+=argp[e].u.vec->size;
      }
      if(num_arg==2)
      {
	w = do_array_surgery(argp[0].u.vec,argp[1].u.vec,OPER_ADD,T_MAPPING);
      }else{
        v=allocate_array(size);
        values=allocate_array(size);
        for(size=e=0;e<num_arg;e++)
        {
          r=argp[e].u.vec->item[0].u.vec->item;
          rr=argp[e].u.vec->item[1].u.vec->item;
          for(d=0;d<argp[e].u.vec->size;d++)
          {
            assign_svalue_no_free(v->item+size,r+d);
            assign_svalue_no_free(values->item+size,rr+d);
            size++;
          }
        }
        w=allocate_array(2);
        w->item[0].u.vec=v;
        w->item[0].type=T_ALIST_PART;
        w->item[1].u.vec=values;
        w->item[1].type=T_ALIST_PART;
        order_alist(w);
      }
      pop_n_elems(num_arg);
      push_mapping(w);
      w->ref--;
    }
    break;
  }
}

INLINE void f_add() { f_sum(2,sp-1); }

void f_subtract()
{
  if ((sp-1)->type != sp->type )
    error("Subtract on different types.\n");

  switch(sp->type)
  {
  case T_POINTER:
  {
    extern struct vector *subtract_array
      PROT((struct vector *,struct vector*));
    struct vector *v;

    /* subtract_array already takes care of destructed objects */
    check_vector_for_destruct(sp[-1].u.vec);
    check_vector_for_destruct(sp[0].u.vec);
    if(sp[0].u.vec->types & sp[-1].u.vec->types)
    {
      v = subtract_array((sp-1)->u.vec, sp->u.vec);
    }else{
      v=slice_array(sp[-1].u.vec,0,sp[-1].u.vec->size-1);
    }
    pop_stack();
    pop_stack();
    push_vector(v);
    v->ref--;
    return;
  }
  case T_LIST:
  case T_MAPPING:
  {
    struct vector *v;
    int type=sp->type;

    v=do_array_surgery((sp-1)->u.vec,sp->u.vec,OPER_SUB,type);
    pop_stack();
    pop_stack();
    push_vector(v);
    v->ref--;
    sp->type=type;
    return;
  }
  case T_FLOAT:
    sp--;
    sp->u.fnum=sp[0].u.fnum-sp[1].u.fnum;
    return;

  case T_NUMBER:
    sp--;
    sp->u.number = sp[0].u.number - sp[1].u.number;
    return;

  case T_STRING:
    push_string("",STRING_CONSTANT);
    f_replace(3,sp-2);
    return;

  default:
    bad_arg(1,F_SUBTRACT);
  }
}

void f_and()
{
  if(sp->type == (sp-1)->type && 
     IS_TYPE(*sp,BT_POINTER | BT_LIST | BT_MAPPING))
   {
    struct vector *v;
    int type=sp->type;

    check_vector_for_destruct(sp[-1].u.vec);
    check_vector_for_destruct(sp[0].u.vec);
    if(sp[-1].u.vec->types & sp[0].u.vec->types)
    {
      v=do_array_surgery((sp-1)->u.vec,sp->u.vec,OPER_AND,type);
    }else{
      v=allocate_array(0);
    }
    pop_stack();
    pop_stack();
    push_vector(v);
    v->ref--;
    sp->type=type;
    return;
  }

  if ((sp-1)->type != T_NUMBER) bad_arg(1,F_AND);
  if (sp->type != T_NUMBER) bad_arg(2,F_AND);
  sp--;
  sp->u.number = sp[0].u.number & sp[1].u.number;
}

void f_or()
{
  if(sp->type == (sp-1)->type &&
     IS_TYPE(*sp,BT_POINTER | BT_LIST | BT_MAPPING))
  {
    struct vector *v;
    int type=sp->type;

    check_vector_for_destruct(sp[-1].u.vec);
    check_vector_for_destruct(sp[0].u.vec);
    if(sp[-1].u.vec->types & sp[0].u.vec->types)
    {
      v=do_array_surgery((sp-1)->u.vec,sp->u.vec,OPER_OR,type);
    }else{
      f_add();
      return;
    }
    pop_stack();
    pop_stack();
    push_vector(v);
    v->ref--;
    sp->type=type;
    return;
  }

  if ((sp-1)->type != T_NUMBER) bad_arg(1,F_OR);
  if (sp->type != T_NUMBER) bad_arg(2,F_OR);
  sp--;
  sp->u.number = sp[0].u.number | sp[1].u.number;
}

void f_xor()
{
  if(sp->type == (sp-1)->type &&
     IS_TYPE(*sp,BT_POINTER | BT_LIST | BT_MAPPING ))
  {
    struct vector *v;
    int type=sp->type;

    check_vector_for_destruct(sp[-1].u.vec);
    check_vector_for_destruct(sp[0].u.vec);
    if(sp[-1].u.vec->types & sp[0].u.vec->types)
    {
      v=do_array_surgery((sp-1)->u.vec,sp->u.vec,OPER_XOR,type);
    }else{
      f_add();
      return;
    }
    pop_stack();
    pop_stack();
    push_vector(v);
    v->ref--;
    sp->type=type;
    return;
  }

  if ((sp-1)->type != T_NUMBER) bad_arg(1,F_XOR);
  if (sp->type != T_NUMBER) bad_arg(2,F_XOR);
  sp--;
  sp->u.number = sp[0].u.number ^ sp[1].u.number;
}

void f_lsh()
{
  if ((sp-1)->type != T_NUMBER) bad_arg(1,F_LSH);
  if (sp->type != T_NUMBER) bad_arg(2,F_LSH);
  sp--;
  sp->u.number = sp[0].u.number << sp[1].u.number;
}

void f_rsh()
{
  if ((sp-1)->type != T_NUMBER) bad_arg(1,F_RSH);
  if (sp->type != T_NUMBER) bad_arg(2,F_RSH);
  sp--;
  sp->u.number = sp[0].u.number >> sp[1].u.number;
}

void f_multiply()
{
  switch(sp[-1].type)
  {
  case T_POINTER:
    if(sp->type!=T_STRING) bad_arg(2,F_MULTIPLY);
    f_implode(2,sp-1);
    return;

  case T_FLOAT:
    if(sp->type!=T_FLOAT) bad_arg(2,F_MULTIPLY);
    sp--;
    sp->u.fnum=sp[0].u.fnum*sp[1].u.fnum;
    return;

  case T_INT:
    if(sp->type!=T_INT) bad_arg(2,F_MULTIPLY);
    sp--;
    sp->u.number=sp[0].u.number*sp[1].u.number;
    return;

  default:
    bad_arg(1,F_MULTIPLY);
    
  }
}

void f_divide()
{
  if(sp[-1].type!=sp[0].type) bad_arg(2,F_DIVIDE);
  switch(sp[0].type)
  {
  case T_STRING:
    f_explode(2,sp-1);
    return;

  case T_FLOAT:
    if(sp->u.fnum==0.0)
      error("Division by zero.\n");
    sp--;
    sp->u.fnum = sp[0].u.fnum / sp[1].u.fnum;
    return;

  case T_NUMBER:
    if (sp->u.number == 0)
      error("Division by zero\n");
    sp--;
    sp->u.number = sp[0].u.number / sp[1].u.number;
    return;
    
  default:
    bad_arg(1,F_DIVIDE);
  }
}

void f_mod()
{
  if(sp[-1].type!=sp[0].type) bad_arg(2,F_MOD);

  if (sp->type == T_FLOAT)
  {
    float foo;
    if(sp->u.fnum==0.0)
      error("Modulo by zero.\n");
    foo=(sp-1)->u.fnum / sp->u.fnum;
    foo=(sp-1)->u.fnum-sp->u.fnum*(float)((int)floor(foo));
    sp--;
    sp->u.fnum=foo;
    return;
  }
  if (sp->type != T_NUMBER) bad_arg(1,F_MOD);
  if (sp->u.number == 0)
    error("Modulus by zero.\n");
  sp--;
  sp->u.number = sp[0].u.number % sp[1].u.number;
}

void f_not()
{
  if(sp->type==T_FLOAT)
  {
    pop_push_conditional(sp->u.fnum==0.0);
    return;
  } 
  pop_push_conditional(sp->type == T_NUMBER && sp->u.number == 0);
}

void f_compl()
{
  if (sp->type != T_NUMBER)
    error("Bad argument to ~\n");
  sp->u.number = ~ sp->u.number;
}

void f_negate()
{
  if(sp->type==T_FLOAT)
  {
    sp->u.fnum=-sp->u.fnum;
    return;
  }
  if (sp->type != T_NUMBER)
    error("Bad argument to unary minus\n");
  sp->u.number = - sp->u.number;
}

void f_inc()
{
  if (sp->type != T_LVALUE)
    error("Bad argument to ++\n");
  if(sp->u.lvalue->sval)
  {
    if (sp->u.lvalue->sval->type != T_NUMBER)
      error("++ of non-numeric argument\n");
    sp->u.lvalue->sval->u.number++;
    lvalue_to_svalue(sp,sp->u.lvalue);
  }else{
    sp++;
    lvalue_to_svalue_no_free(sp,sp[-1].u.lvalue);
    push_number(1);
    f_add();
    f_assign();
  }
}
void f_inc_and_pop() { f_inc(); pop_stack(); }

void f_dec()
{
  if (sp->type != T_LVALUE)
    error("Bad argument to --\n");
  if(sp->u.lvalue->sval)
  {
    if (sp->u.lvalue->sval->type != T_NUMBER)
      error("-- of non-numeric argument\n");
    sp->u.lvalue->sval->u.number--;
    lvalue_to_svalue(sp,sp->u.lvalue);
  }else{
    sp++;
    lvalue_to_svalue_no_free(sp,sp[-1].u.lvalue);
    push_number(-1);
    f_add();
    f_assign();
  }
}

void f_dec_and_pop() { f_dec(); pop_stack(); }
 

void f_post_inc()
{
  if (sp->type != T_LVALUE)
    error("Bad argument to ++\n");
  if(sp->u.lvalue->sval)
  {
    if (sp->u.lvalue->sval->type != T_NUMBER)
      error("++ of non-numeric argument\n");
    sp->u.lvalue->sval->u.number++;
    lvalue_to_svalue(sp,sp->u.lvalue);
    sp->u.number--;
  }else{
    sp++;
    lvalue_to_svalue_no_free(sp,sp[-1].u.lvalue);
    f_swap();
    sp++;
    assign_svalue_no_free(sp,sp-2);
    push_number(1);
    f_add();
    f_assign_and_pop();
  }
}

void f_post_dec()
{
  if (sp->type != T_LVALUE)
    error("Bad argument to --\n");
  if(sp->u.lvalue->sval)
  {
    if (sp->u.lvalue->sval->type != T_NUMBER)
      error("-- of non-numeric argument\n");
    sp->u.lvalue->sval->u.number--;
    lvalue_to_svalue(sp,sp->u.lvalue);
    sp->u.number++;
  }else{
    sp++;
    lvalue_to_svalue_no_free(sp,sp[-1].u.lvalue);
    f_swap();
    sp++;
    assign_svalue_no_free(sp,sp-2);
    push_number(-1);
    f_add();
    f_assign_and_pop();
  }
}

void f_assign()
{
#ifdef DEBUG
  if (sp[-1].type != T_LVALUE)
    error("Bad argument to F_ASSIGN\n");
#endif
  assign(sp[-1].u.lvalue,sp);
  free_svalue(sp-1);
  sp[-1]=*sp;
  *sp=const0;
  sp--;
}

void f_assign_and_pop()
{
#ifdef DEBUG
  if (sp[-1].type != T_LVALUE)
    error("Bad argument to F_ASSIGN\n");
#endif
  assign(sp[-1].u.lvalue,sp);
  pop_stack();
  pop_stack();
}

void f_is_equal(int num_arg,struct svalue *argp)
{
  int i;
  i=is_equal(argp,argp+1);
  pop_stack();
  pop_stack();
  push_number(i);
}
